@@ -1,7 +1,7 @@
 /*
  * relay-irc.c - IRC protocol for relay to client: IRC proxy/bouncer
  *
- * Copyright (C) 2003-2022 Sébastien Helleu <flashcode@flashtux.org>
+ * Copyright (C) 2003-2023 Sébastien Helleu <flashcode@flashtux.org>
  *
  * This file is part of WeeChat, the extensible chat client.
  *
@@ -43,7 +43,7 @@ char *relay_irc_ignore_commands[] = { "cap", "pong", "quit", NULL };
 char *relay_irc_backlog_commands_tags[RELAY_IRC_NUM_CMD] =
 { "irc_join", "irc_part", "irc_quit", "irc_nick", "irc_privmsg" };
 char *relay_irc_server_capabilities[RELAY_IRC_NUM_CAPAB] =
-{ "server-time" };
+{ "server-time", "echo-message" };
 
 
 /*
@@ -59,13 +59,13 @@ relay_irc_command_relayed (const char *irc_command)
 {
     int i;
 
-    if (irc_command)
+    if (!irc_command)
+        return 0;
+
+    for (i = 0; relay_irc_relay_commands[i]; i++)
     {
-        for (i = 0; relay_irc_relay_commands[i]; i++)
-        {
-            if (weechat_strcasecmp (relay_irc_relay_commands[i], irc_command) == 0)
-                return 1;
-        }
+        if (weechat_strcasecmp (relay_irc_relay_commands[i], irc_command) == 0)
+            return 1;
     }
 
     /* command must NOT be relayed to client */
@@ -85,16 +85,16 @@ relay_irc_command_ignored (const char *irc_command)
 {
     int i;
 
-    if (irc_command)
+    if (!irc_command)
+        return 0;
+
+    for (i = 0; relay_irc_ignore_commands[i]; i++)
     {
-        for (i = 0; relay_irc_ignore_commands[i]; i++)
-        {
-            if (weechat_strcasecmp (relay_irc_ignore_commands[i], irc_command) == 0)
-                return 1;
-        }
+        if (weechat_strcasecmp (relay_irc_ignore_commands[i], irc_command) == 0)
+            return 1;
     }
 
-    /* command must NOT be relayed to client */
+    /* command must be ignored */
     return 0;
 }
 
@@ -109,6 +109,9 @@ relay_irc_search_backlog_commands_tags (const char *tag)
 {
     int i;
 
+    if (!tag)
+        return -1;
+
     for (i = 0; i < RELAY_IRC_NUM_CMD; i++)
     {
         if (strcmp (relay_irc_backlog_commands_tags[i], tag) == 0)
@@ -131,6 +134,9 @@ relay_irc_search_server_capability (const char *capability)
 {
     int i;
 
+    if (!capability)
+        return -1;
+
     for (i = 0; i < RELAY_IRC_NUM_CAPAB; i++)
     {
         if (strcmp (relay_irc_server_capabilities[i], capability) == 0)
@@ -152,6 +158,9 @@ relay_irc_message_parse (const char *message)
 {
     struct t_hashtable *hash_msg, *hash_parsed;
 
+    if (!message)
+        return NULL;
+
     hash_msg = NULL;
     hash_parsed = NULL;
 
@@ -192,8 +201,9 @@ void
 relay_irc_sendf (struct t_relay_client *client, const char *format, ...)
 {
     int length, number;
-    char *pos, hash_key[32], *message;
-    const char *str_message;
+    char *pos, hash_key[32], *message, *new_msg1, *new_msg2;
+    char modifier_data[128];
+    const char *str_message, *ptr_msg1, *ptr_msg2;
     struct t_hashtable *hashtable_in, *hashtable_out;
 
     if (!client)
@@ -203,10 +213,32 @@ relay_irc_sendf (struct t_relay_client *client, const char *format, ...)
     if (!vbuffer)
         return;
 
-    pos = strchr (vbuffer, '\r');
+    new_msg1 = NULL;
+    hashtable_in = NULL;
+    hashtable_out = NULL;
+
+    snprintf (modifier_data, sizeof (modifier_data),
+              "0x%lx", (unsigned long)client);
+    new_msg1 = weechat_hook_modifier_exec ("relay_client_irc_out1",
+                                           modifier_data, vbuffer);
+
+    /* no changes in new message? */
+    if (new_msg1 && (strcmp (vbuffer, new_msg1) == 0))
+    {
+        free (new_msg1);
+        new_msg1 = NULL;
+    }
+
+    /* message dropped? */
+    if (new_msg1 && !new_msg1[0])
+        goto end;
+
+    ptr_msg1 = (new_msg1) ? new_msg1 : vbuffer;
+
+    pos = strchr (ptr_msg1, '\r');
     if (pos)
         pos[0] = '\0';
-    pos = strchr (vbuffer, '\n');
+    pos = strchr (ptr_msg1, '\n');
     if (pos)
         pos[0] = '\0';
 
@@ -214,40 +246,100 @@ relay_irc_sendf (struct t_relay_client *client, const char *format, ...)
                                           WEECHAT_HASHTABLE_STRING,
                                           WEECHAT_HASHTABLE_STRING,
                                           NULL, NULL);
-    if (hashtable_in)
+    if (!hashtable_in)
+        goto end;
+
+    weechat_hashtable_set (hashtable_in, "server", client->protocol_args);
+    weechat_hashtable_set (hashtable_in, "message", ptr_msg1);
+    hashtable_out = weechat_info_get_hashtable ("irc_message_split",
+                                                hashtable_in);
+    if (!hashtable_out)
+        goto end;
+
+    number = 1;
+    while (1)
     {
-        weechat_hashtable_set (hashtable_in, "server", client->protocol_args);
-        weechat_hashtable_set (hashtable_in, "message", vbuffer);
-        hashtable_out = weechat_info_get_hashtable ("irc_message_split",
-                                                    hashtable_in);
-        if (hashtable_out)
+        snprintf (hash_key, sizeof (hash_key), "msg%d", number);
+        str_message = weechat_hashtable_get (hashtable_out, hash_key);
+        if (!str_message)
+            break;
+        new_msg2 = weechat_hook_modifier_exec ("relay_client_irc_out",
+                                               modifier_data, str_message);
+        /* no changes in new message? */
+        if (new_msg2 && (strcmp (str_message, new_msg2) == 0))
         {
-            number = 1;
-            while (1)
-            {
-                snprintf (hash_key, sizeof (hash_key), "msg%d", number);
-                str_message = weechat_hashtable_get (hashtable_out, hash_key);
-                if (!str_message)
-                    break;
-                length = strlen (str_message) + 16 + 1;
-                message = malloc (length);
-                if (message)
-                {
-                    snprintf (message, length, "%s\r\n", str_message);
-                    relay_client_send (client, RELAY_CLIENT_MSG_STANDARD,
-                                       message, strlen (message), NULL);
-                    free (message);
-                }
-                number++;
-            }
-            weechat_hashtable_free (hashtable_out);
+            free (new_msg2);
+            new_msg2 = NULL;
         }
-        weechat_hashtable_free (hashtable_in);
+        /* message not dropped? */
+        if (!new_msg2 || new_msg2[0])
+        {
+            ptr_msg2 = (new_msg2) ? new_msg2 : str_message;
+            length = strlen (ptr_msg2) + 16 + 1;
+            message = malloc (length);
+            if (message)
+            {
+                snprintf (message, length, "%s\r\n", ptr_msg2);
+                relay_client_send (client, RELAY_CLIENT_MSG_STANDARD,
+                                   message, strlen (message), NULL);
+                free (message);
+            }
+        }
+        if (new_msg2)
+            free (new_msg2);
+        number++;
     }
 
+end:
+    if (new_msg1)
+        free (new_msg1);
+    if (hashtable_in)
+        weechat_hashtable_free (hashtable_in);
+    if (hashtable_out)
+        weechat_hashtable_free (hashtable_out);
     free (vbuffer);
 }
 
+/*
+ * Parses CAP command received from IRC server.
+ */
+
+void
+relay_irc_parse_cap_message (struct t_relay_client *client,
+                             struct t_hashtable *parsed_msg)
+{
+    const char *ptr_param;
+    char str_param[64], **caps;
+    int i, index, num_caps;
+
+    /* only CAP ACK is parsed */
+    ptr_param = weechat_hashtable_get (parsed_msg, "param2");
+    if (!ptr_param || (weechat_strcasecmp (ptr_param, "ACK") != 0))
+        return;
+
+    index = 3;
+    while (1)
+    {
+        snprintf (str_param, sizeof (str_param), "param%d", index);
+        ptr_param = weechat_hashtable_get (parsed_msg, str_param);
+        if (!ptr_param)
+            break;
+        caps = weechat_string_split (ptr_param, " ", NULL, 0, 0, &num_caps);
+        if (caps)
+        {
+            for (i = 0; i < num_caps; i++)
+            {
+                if (strcmp (caps[i], "-echo-message") == 0)
+                    RELAY_IRC_DATA(client, irc_cap_echo_message) = 0;
+                else if (strcmp (caps[i], "echo-message") == 0)
+                    RELAY_IRC_DATA(client, irc_cap_echo_message) = 1;
+            }
+            weechat_string_free_split (caps);
+        }
+        index++;
+    }
+}
+
 /*
  * Callback for signal "irc_in2".
  *
@@ -302,6 +394,12 @@ relay_irc_signal_irc_in2_cb (const void *pointer, void *data,
                                                    irc_args + 1 : irc_args);
         }
 
+        /* if capabilities have changed, parse them to update in client */
+        if (irc_command && (weechat_strcasecmp (irc_command, "cap") == 0))
+        {
+            relay_irc_parse_cap_message (client, hash_parsed);
+        }
+
         /* relay all commands to client, but not ping/pong */
         if (irc_command
             && (weechat_strcasecmp (irc_command, "ping") != 0)
@@ -335,30 +433,30 @@ relay_irc_tag_relay_client_id (const char *tags)
 
     result = -1;
 
-    if (tags && tags[0])
+    if (!tags || !tags[0])
+        return result;
+
+    argv = weechat_string_split (tags, ",", NULL,
+                                 WEECHAT_STRING_SPLIT_STRIP_LEFT
+                                 | WEECHAT_STRING_SPLIT_STRIP_RIGHT
+                                 | WEECHAT_STRING_SPLIT_COLLAPSE_SEPS,
+                                 0, &argc);
+    if (argv)
     {
-        argv = weechat_string_split (tags, ",", NULL,
-                                     WEECHAT_STRING_SPLIT_STRIP_LEFT
-                                     | WEECHAT_STRING_SPLIT_STRIP_RIGHT
-                                     | WEECHAT_STRING_SPLIT_COLLAPSE_SEPS,
-                                     0, &argc);
-        if (argv)
+        for (i = 0; i < argc; i++)
         {
-            for (i = 0; i < argc; i++)
+            if (strncmp (argv[i], "relay_client_", 13) == 0)
             {
-                if (strncmp (argv[i], "relay_client_", 13) == 0)
+                error = NULL;
+                number = strtol (argv[i] + 13, &error, 10);
+                if (error && !error[0])
                 {
-                    error = NULL;
-                    number = strtol (argv[i] + 13, &error, 10);
-                    if (error && !error[0])
-                    {
-                        result = number;
-                        break;
-                    }
+                    result = number;
+                    break;
                 }
             }
-            weechat_string_free_split (argv);
         }
+        weechat_string_free_split (argv);
     }
 
     return result;
@@ -444,29 +542,36 @@ relay_irc_signal_irc_outtags_cb (const void *pointer, void *data,
             && irc_channel && irc_channel[0]
             && relay_irc_command_relayed (irc_command))
         {
-            /* get host for nick (it is self nick) */
-            snprintf (str_infolist_args, sizeof (str_infolist_args),
-                      "%s,%s,%s",
-                      client->protocol_args,
-                      irc_channel,
-                      RELAY_IRC_DATA(client, nick));
+            /*
+             * relay command only if capability echo-message is NOS enabled
+             * in IRC server (otherwise message would be displayed two times)
+             */
+            if (!RELAY_IRC_DATA(client, irc_cap_echo_message))
+            {
+                /* get host for nick (it is self nick) */
+                snprintf (str_infolist_args, sizeof (str_infolist_args),
+                          "%s,%s,%s",
+                          client->protocol_args,
+                          irc_channel,
+                          RELAY_IRC_DATA(client, nick));
 
-            host = NULL;
-            infolist_nick = weechat_infolist_get ("irc_nick", NULL,
-                                                  str_infolist_args);
-            if (infolist_nick && weechat_infolist_next (infolist_nick))
-                host = weechat_infolist_string (infolist_nick, "host");
+                host = NULL;
+                infolist_nick = weechat_infolist_get ("irc_nick", NULL,
+                                                      str_infolist_args);
+                if (infolist_nick && weechat_infolist_next (infolist_nick))
+                    host = weechat_infolist_string (infolist_nick, "host");
 
-            /* send message to client */
-            relay_irc_sendf (client,
-                             ":%s%s%s %s",
-                             RELAY_IRC_DATA(client, nick),
-                             (host && host[0]) ? "!" : "",
-                             (host && host[0]) ? host : "",
-                             ptr_message);
+                /* send message to client */
+                relay_irc_sendf (client,
+                                 ":%s%s%s %s",
+                                 RELAY_IRC_DATA(client, nick),
+                                 (host && host[0]) ? "!" : "",
+                                 (host && host[0]) ? host : "",
+                                 ptr_message);
 
-            if (infolist_nick)
-                weechat_infolist_free (infolist_nick);
+                if (infolist_nick)
+                    weechat_infolist_free (infolist_nick);
+            }
         }
         if (irc_channel)
             free (irc_channel);
@@ -1140,7 +1245,7 @@ relay_irc_send_join_channels (struct t_relay_client *client)
 
 void
 relay_irc_input_send (struct t_relay_client *client, const char *irc_channel,
-                      char *options, const char *format, ...)
+                      const char *options, const char *format, ...)
 {
     char buf_beginning[1024], *buf;
     int length_beginning, length_vbuffer;
@@ -1233,6 +1338,125 @@ relay_irc_hook_signals (struct t_relay_client *client)
                               client, NULL);
 }
 
+/*
+ * Compares two capabilities.
+ *
+ * Returns:
+ *   < 0: capability 1 < capability 2
+ *     0: capability 1 == capability 2
+ *   > 0: capability 1 > capability 2
+ */
+
+int
+relay_irc_capability_compare_cb (void *data,
+                                 struct t_arraylist *arraylist,
+                                 void *pointer1,
+                                 void *pointer2)
+{
+    /* make C compiler happy */
+    (void) data;
+    (void) arraylist;
+
+    return (weechat_strcmp ((const char *)pointer1, (const char *)pointer2));
+}
+
+/*
+ * Frees a capability in list.
+ */
+
+void
+relay_irc_capability_free_db (void *data,
+                              struct t_arraylist *arraylist,
+                              void *pointer)
+{
+    /* make C compiler happy */
+    (void) data;
+    (void) arraylist;
+
+    free (pointer);
+}
+
+/*
+ * Checks if a capability is currently enabled in IRC server.
+ *
+ * Returns:
+ *   1: capability is enabled in server
+ *   0: capability is NOT enabled in server
+ */
+
+int
+relay_irc_cap_enabled (struct t_relay_client *client, const char *capability)
+{
+    char str_info[1024], *info;
+    int rc;
+
+    if (!client || !capability || !capability[0])
+        return 0;
+
+    snprintf (str_info, sizeof (str_info),
+              "%s,%s",
+              client->protocol_args,
+              capability);
+
+    info = weechat_info_get ("irc_server_cap", str_info);
+    rc = (info && (strcmp (info, "1") == 0)) ? 1 : 0;
+    if (info)
+        free (info);
+    return rc;
+}
+
+/*
+ * Returns an integer with bits set for each supported capability (ie that
+ * is either always supported, or enabled in the IRC server).
+ */
+
+int
+relay_irc_get_supported_caps (struct t_relay_client *client)
+{
+    int i, caps, check_server;
+
+    caps = 0;
+    for (i = 0; i < RELAY_IRC_NUM_CAPAB; i++)
+    {
+        check_server = RELAY_IRC_CAPAB_FOLLOW_SERVER & (1 << i);
+        if (!check_server
+            || relay_irc_cap_enabled (client, relay_irc_server_capabilities[i]))
+        {
+            caps |= (1 << i);
+        }
+    }
+
+    return caps;
+}
+
+/*
+ * Returns a sorted list of supported all server capabilities by relay (even
+ * those that should not be sent to clients when the IRC server doesn't
+ * support them).
+ *
+ * Note: result must be freed after use.
+ */
+
+struct t_arraylist *
+relay_irc_get_list_caps ()
+{
+    struct t_arraylist *list_capab;
+    int i;
+
+    list_capab = weechat_arraylist_new (
+        8, 1, 0,
+        &relay_irc_capability_compare_cb, NULL,
+        &relay_irc_capability_free_db, NULL);
+
+    for (i = 0; i < RELAY_IRC_NUM_CAPAB; i++)
+    {
+        weechat_arraylist_add (list_capab,
+                               strdup (relay_irc_server_capabilities[i]));
+    }
+
+    return list_capab;
+}
+
 /*
  * Processes the "CAP" irc command (received from client)
  */
@@ -1241,8 +1465,11 @@ void
 relay_irc_recv_command_capab (struct t_relay_client *client,
                               int num_params, const char **params)
 {
-    char str_capab[1024], *str_caps;
-    int i, capability, server_caps, num_caps_received, caps_ok;
+    struct t_arraylist *list_caps;
+    char **str_caps, **caps;
+    const char *ptr_cap;
+    int i, j, capability, server_caps, num_caps_received, caps_ok, size;
+    int num_caps, supported_caps;
 
     if (num_params < 1)
         return;
@@ -1250,17 +1477,37 @@ relay_irc_recv_command_capab (struct t_relay_client *client,
     if (weechat_strcasecmp (params[0], "ls") == 0)
     {
         /* return the list of supported server capabilities */
-        str_capab[0] = '\0';
-        for (i = 0; i < RELAY_IRC_NUM_CAPAB; i++)
+        list_caps = relay_irc_get_list_caps ();
+        if (list_caps)
         {
-            if (str_capab[0])
-                strcat (str_capab, " ");
-            strcat (str_capab, relay_irc_server_capabilities[i]);
-            relay_irc_sendf (client,
-                             ":%s CAP %s LS :%s",
-                             RELAY_IRC_DATA(client, address),
-                             (RELAY_IRC_DATA(client, nick)) ? RELAY_IRC_DATA(client, nick) : "nick",
-                             str_capab);
+            supported_caps = relay_irc_get_supported_caps (client);
+            str_caps = weechat_string_dyn_alloc (256);
+            if (str_caps)
+            {
+                size = weechat_arraylist_size (list_caps);
+                for (i = 0; i < size; i++)
+                {
+                    ptr_cap = (const char *)weechat_arraylist_get (list_caps, i);
+                    capability = relay_irc_search_server_capability (ptr_cap);
+                    if (capability >= 0)
+                    {
+                        if (supported_caps & (1 << capability))
+                        {
+                            if ((*str_caps)[0])
+                                weechat_string_dyn_concat (str_caps, " ", -1);
+                            weechat_string_dyn_concat (str_caps, ptr_cap, -1);
+                        }
+                    }
+                }
+                relay_irc_sendf (
+                    client,
+                    ":%s CAP %s LS :%s",
+                    RELAY_IRC_DATA(client, address),
+                    (RELAY_IRC_DATA(client, nick)) ? RELAY_IRC_DATA(client, nick) : "nick",
+                    *str_caps);
+                weechat_string_dyn_free (str_caps, 1);
+            }
+            weechat_arraylist_free (list_caps);
         }
         if (!RELAY_IRC_DATA(client, connected))
             RELAY_IRC_DATA(client, cap_ls_received) = 1;
@@ -1268,47 +1515,63 @@ relay_irc_recv_command_capab (struct t_relay_client *client,
     else if (weechat_strcasecmp (params[0], "req") == 0)
     {
         /* client is asking for one or more server capabilities */
-        num_caps_received = 0;
-        caps_ok = 0;
-        server_caps = RELAY_IRC_DATA(client, server_capabilities);
-        for (i = 1; i < num_params; i++)
+        list_caps = relay_irc_get_list_caps ();
+        if (list_caps)
         {
-            num_caps_received++;
-            capability = relay_irc_search_server_capability (params[i]);
-            if (capability >= 0)
+            supported_caps = relay_irc_get_supported_caps (client);
+            num_caps_received = 0;
+            caps_ok = 1;
+            server_caps = RELAY_IRC_DATA(client, server_capabilities);
+            str_caps = weechat_string_dyn_alloc (256);
+            if (str_caps)
             {
-                caps_ok = 1;
-                server_caps |= 1 << capability;
+                for (i = 1; i < num_params; i++)
+                {
+                    if (!params[i][0])
+                        continue;
+                    if ((*str_caps)[0])
+                        weechat_string_dyn_concat (str_caps, " ", -1);
+                    weechat_string_dyn_concat (str_caps, params[i], -1);
+                    caps = weechat_string_split (params[i], " ", NULL, 0, 0,
+                                                 &num_caps);
+                    if (caps)
+                    {
+                        for (j = 0; j < num_caps; j++)
+                        {
+                            num_caps_received++;
+                            capability = relay_irc_search_server_capability (caps[j]);
+                            if ((capability >= 0) && (supported_caps & (1 << capability)))
+                                server_caps |= 1 << capability;
+                            else
+                                caps_ok = 0;
+                        }
+                        weechat_string_free_split (caps);
+                    }
+                }
+                if (num_caps_received == 0)
+                    caps_ok = 0;
+                if (caps_ok)
+                    RELAY_IRC_DATA(client, server_capabilities) = server_caps;
+                relay_irc_sendf (
+                    client,
+                    ":%s CAP %s %s :%s",
+                    RELAY_IRC_DATA(client, address),
+                    (RELAY_IRC_DATA(client, nick)) ? RELAY_IRC_DATA(client, nick) : "nick",
+                    (caps_ok) ? "ACK" : "NAK",
+                    *str_caps);
+                weechat_string_dyn_free (str_caps, 1);
             }
-            else
+            /*
+             * if the CAP REQ command is received without arguments, we consider
+             * the CAP END is received; this is a workaround for clients like
+             * Atomic which are sending "CAP REQ :" (see issue #1040)
+             */
+            if (num_caps_received == 0)
             {
-                caps_ok = 0;
-                break;
+                if (!RELAY_IRC_DATA(client, connected))
+                    RELAY_IRC_DATA(client, cap_end_received) = 1;
             }
-        }
-        if (caps_ok)
-            RELAY_IRC_DATA(client, server_capabilities) = server_caps;
-        str_caps = (num_params > 1) ?
-            weechat_string_rebuild_split_string (params, " ", 1, -1) : NULL;
-        relay_irc_sendf (
-            client,
-            ":%s CAP %s %s :%s",
-            RELAY_IRC_DATA(client, address),
-            (RELAY_IRC_DATA(client, nick)) ? RELAY_IRC_DATA(client, nick) : "nick",
-            (caps_ok) ? "ACK" : "NAK",
-            (str_caps) ? str_caps : "");
-        if (str_caps)
-            free (str_caps);
-
-        /*
-         * if the CAP REQ command is received without arguments, we consider
-         * the CAP END is received; this is a workaround for clients like
-         * Atomic which are sending "CAP REQ :" (see issue #1040)
-         */
-        if (num_caps_received == 0)
-        {
-            if (!RELAY_IRC_DATA(client, connected))
-                RELAY_IRC_DATA(client, cap_end_received) = 1;
+            weechat_arraylist_free (list_caps);
         }
     }
     else if (weechat_strcasecmp (params[0], "end") == 0)
@@ -1328,12 +1591,15 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
     struct t_hashtable *hash_parsed, *hash_redirect;
     struct t_infolist *infolist_server;
     const char *irc_command, *str_num_params, *isupport, *pos_password;
+    const char *ptr_data;
     char str_time[128], str_signal[128], str_server_channel[256], *nick;
     char str_param[128], *str_args, *version, str_command[128], **params;
-    char *pos, *password, *irc_is_channel, *info, *error;
+    char *pos, *password, *irc_is_channel, *info, *error, *str_cmd_lower;
+    char modifier_data[128], *new_data;
     long num_params;
     int i, redirect_msg;
 
+    new_data = NULL;
     hash_parsed = NULL;
     params = NULL;
 
@@ -1348,8 +1614,26 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
                         data);
     }
 
+    snprintf (modifier_data, sizeof (modifier_data),
+              "0x%lx", (unsigned long)client);
+    new_data = weechat_hook_modifier_exec ("relay_client_irc_in",
+                                           modifier_data, data);
+
+    /* no changes in new data */
+    if (new_data && (strcmp (data, new_data) == 0))
+    {
+        free (new_data);
+        new_data = NULL;
+    }
+
+    /* message dropped? */
+    if (new_data && !new_data[0])
+        goto end;
+
+    ptr_data = (new_data) ? new_data : data;
+
     /* parse IRC message */
-    hash_parsed = relay_irc_message_parse (data);
+    hash_parsed = relay_irc_message_parse (ptr_data);
     if (!hash_parsed)
         goto end;
     irc_command = weechat_hashtable_get (hash_parsed, "command");
@@ -1387,11 +1671,7 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
     /* server capabilities */
     if (irc_command && (weechat_strcasecmp (irc_command, "cap") == 0))
     {
-        if (num_params > 0)
-        {
-            relay_irc_recv_command_capab (client,
-                                          num_params, (const char **)params);
-        }
+        relay_irc_recv_command_capab (client, num_params, (const char **)params);
     }
     /* if client is not yet "connected" */
     if (!RELAY_IRC_DATA(client, connected))
@@ -1469,6 +1749,7 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
                                          RELAY_IRC_DATA(client, address));
                         relay_client_set_status (client,
                                                  RELAY_STATUS_DISCONNECTED);
+                        weechat_infolist_free (infolist_server);
                         goto end;
                     }
                     if (num_params > 0)
@@ -1496,6 +1777,8 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
             }
 
             RELAY_IRC_DATA(client, connected) = 1;
+            RELAY_IRC_DATA(client, irc_cap_echo_message) = relay_irc_cap_enabled (
+                client, "echo-message");
 
             /*
              * send nick to client if server nick is different of nick asked
@@ -1766,11 +2049,15 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
                         redirect_msg = 1;
                         snprintf (str_command, sizeof (str_command),
                                   "%s", irc_command);
-                        weechat_string_tolower (str_command);
-                        weechat_hashtable_set (hash_redirect, "pattern",
-                                               str_command);
-                        weechat_hashtable_set (hash_redirect, "string",
-                                               params[0]);
+                        str_cmd_lower = weechat_string_tolower (str_command);
+                        if (str_cmd_lower)
+                        {
+                            weechat_hashtable_set (hash_redirect, "pattern",
+                                                   str_cmd_lower);
+                            weechat_hashtable_set (hash_redirect, "string",
+                                                   params[0]);
+                            free (str_cmd_lower);
+                        }
                     }
                 }
                 else if (weechat_strcasecmp (irc_command, "time") == 0)
@@ -1798,11 +2085,13 @@ relay_irc_recv (struct t_relay_client *client, const char *data)
             relay_irc_input_send (client, NULL,
                                   "priority_high",
                                   "/quote %s",
-                                  data);
+                                  ptr_data);
         }
     }
 
 end:
+    if (new_data)
+        free (new_data);
     if (hash_parsed)
         weechat_hashtable_free (hash_parsed);
     if (params)
@@ -1862,6 +2151,7 @@ relay_irc_alloc (struct t_relay_client *client)
         RELAY_IRC_DATA(client, cap_ls_received) = 0;
         RELAY_IRC_DATA(client, cap_end_received) = 0;
         RELAY_IRC_DATA(client, connected) = 0;
+        RELAY_IRC_DATA(client, irc_cap_echo_message) = 0;
         RELAY_IRC_DATA(client, server_capabilities) = 0;
         RELAY_IRC_DATA(client, hook_signal_irc_in2) = NULL;
         RELAY_IRC_DATA(client, hook_signal_irc_outtags) = NULL;
@@ -1896,6 +2186,7 @@ relay_irc_alloc_with_infolist (struct t_relay_client *client,
         RELAY_IRC_DATA(client, cap_ls_received) = weechat_infolist_integer (infolist, "cap_ls_received");
         RELAY_IRC_DATA(client, cap_end_received) = weechat_infolist_integer (infolist, "cap_end_received");
         RELAY_IRC_DATA(client, connected) = weechat_infolist_integer (infolist, "connected");
+        RELAY_IRC_DATA(client, irc_cap_echo_message) = weechat_infolist_integer (infolist, "irc_cap_echo_message");
         RELAY_IRC_DATA(client, server_capabilities) = weechat_infolist_integer (infolist, "server_capabilities");
         if (RELAY_IRC_DATA(client, connected))
         {
@@ -1996,6 +2287,8 @@ relay_irc_add_to_infolist (struct t_infolist_item *item,
         return 0;
     if (!weechat_infolist_new_var_integer (item, "cap_end_received", RELAY_IRC_DATA(client, cap_end_received)))
         return 0;
+    if (!weechat_infolist_new_var_integer (item, "irc_cap_echo_message", RELAY_IRC_DATA(client, irc_cap_echo_message)))
+        return 0;
     if (!weechat_infolist_new_var_integer (item, "server_capabilities", RELAY_IRC_DATA(client, server_capabilities)))
         return 0;
     if (!weechat_infolist_new_var_pointer (item, "hook_signal_irc_in2", RELAY_IRC_DATA(client, hook_signal_irc_in2)))
@@ -2026,6 +2319,7 @@ relay_irc_print_log (struct t_relay_client *client)
         weechat_log_printf ("    cap_ls_received . . . . : %d",    RELAY_IRC_DATA(client, cap_ls_received));
         weechat_log_printf ("    cap_end_received. . . . : %d",    RELAY_IRC_DATA(client, cap_end_received));
         weechat_log_printf ("    connected . . . . . . . : %d",    RELAY_IRC_DATA(client, connected));
+        weechat_log_printf ("    irc_cap_echo_message. . : %d",    RELAY_IRC_DATA(client, irc_cap_echo_message));
         weechat_log_printf ("    server_capabilities . . : %d",    RELAY_IRC_DATA(client, server_capabilities));
         weechat_log_printf ("    hook_signal_irc_in2 . . : 0x%lx", RELAY_IRC_DATA(client, hook_signal_irc_in2));
         weechat_log_printf ("    hook_signal_irc_outtags : 0x%lx", RELAY_IRC_DATA(client, hook_signal_irc_outtags));
